//
// Copyright (c) 2017 Phobyx GmbH&Co.KG, Gerrit Meyer
//
// This software is provided 'as-is', without any express or implied
// warranty. In no event will the authors be held liable for any damages
// arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it
// freely, subject to the following restrictions:
//
//  1. The origin of this software must not be misrepresented; you must not
//  claim that you wrote the original software. If you use this software
//  in a product, an acknowledgment in the product documentation would be
//  appreciated but is not required.
//
//  2. Altered source versions must be plainly marked as such, and must not be
//  misrepresented as being the original software.
//
//  3. This notice may not be removed or altered from any source
//  distribution.
//
//  4. Any compiled module must not break compatibility to gonuts or be used 
//	for reverse engineering or to circumvent protection and license mechanisms.

#include "sqdbi.h"
#ifndef _SQ64
#include <sstream>
#endif
#ifdef SQUNICODE
#include <cwchar>
#else
#include <cstdio>
#endif
#include <cstring>
SQInteger DBICon::miInstcounter=0 ;

void SQDBI_RegisterClasses( HSQUIRRELVM v )
{
	DBICon::Register(v) ;
	DBIResult::Register(v) ;

}

DBIResult::DBIResult( DBICon *con,dbi_result res,std::string &errmsg )
:mConn(con),mResult(res),m_errmsg(errmsg)
{
	con->AddRef() ;
}

DBIResult::~DBIResult()
{
	if (mConn) mConn->ReleaseRef() ;
	if (mResult) dbi_result_free(mResult) ;
}

void DBIResult::BindScriptMethods( HSQUIRRELVM v)
{
	DENBIND_FUNC( v, constructor ) ;
	DENBIND_FUNC( v, next_row) ;
	DENBIND_FUNC( v, first_row) ;
	DENBIND_FUNC( v, column );
	DENBIND_FUNC( v, table ); //an alias for row
	DENBIND_FUNC( v, row );
	DENBIND_FUNC( v, array );
	DENBIND_FUNC( v, Error) ;
}



static DenObject GetField(HSQUIRRELVM v,dbi_result pxres,SQInteger i)
{
	DenObject NullObj;
	if (dbi_result_field_is_null_idx(pxres,i)) return NullObj ;
	else {
		unsigned short dbitype=dbi_result_get_field_type_idx(pxres,i) ;
		switch (dbitype)
		{
		case DBI_TYPE_INTEGER: 
			{ 
#ifdef _SQ64
				SQInteger ival=(SQInteger)dbi_result_get_ulonglong_idx(pxres,i) ;	
				SQAPI->pushinteger(v,ival) ;
#else
				unsigned long long lval=dbi_result_get_ulonglong_idx(pxres,i) ;
				SQInteger ival=(SQInteger) lval ;
				if (lval==(unsigned long long)ival) 	SQAPI->pushinteger(v,lval) ;	
				else
				{
					SQChar* buf=SQAPI->getscratchpad(v,64) ;
#ifdef SQUNICODE
					swprintf(buf,63,"%lld",lval);
#else	
					snprintf(buf,64,"%lld",lval) ;
#endif
					SQAPI->pushstring(v,buf,-1) ;
				}
#endif				
				DenObject ret(v) ;
				return ret ;
				

			}
		case DBI_TYPE_STRING:   {
								  SQAPI->pushstring(v,dbi_result_get_string_idx(pxres,i),dbi_result_get_field_length_idx(pxres,i)) ; 
								  DenObject ret(v) ;
								  return ret ;
								}

		case DBI_TYPE_DECIMAL: { 
									int attrb=dbi_result_get_field_attribs_idx(pxres,i) ;
									SQFloat fval=0 ;
									if (attrb&DBI_DECIMAL_SIZE8) 
									{ fval=(SQFloat)dbi_result_get_double_idx(pxres,i) ;  }
									else 
									{ fval=(SQFloat)dbi_result_get_float_idx(pxres,i) ; }
									SQAPI->pushfloat(v,fval) ;
									DenObject ret(v) ;
									return ret ;
							   }
		case DBI_TYPE_DATETIME: {
									time_t t = dbi_result_get_datetime_idx( pxres, i ) ;
									SQAPI->pushinteger(v,(SQInteger)t) ;
									DenObject ret(v) ;
									return ret ;
								}
		case DBI_TYPE_BINARY: {
								const unsigned char *data = dbi_result_get_binary_idx( pxres, i ) ; 
								size_t len = dbi_result_get_field_length_idx( pxres, i ) ; 
								if (( data ) && (len)) 
								{
									DENAPI.CreateEmptyBuffer(v,len) ;
									SQChar *buf=DENAPI.GetAnyBuffer(v,-1,NULL) ;
									memcpy( buf, data, len ) ;
									DenObject ret(v) ;
									return ret ;
								}
								return NullObj ;
							  }
		case DBI_TYPE_ERROR: 
					SQAPI->throwerror(v,"column access error") ;
					//deliberately runs into default case:
		default: {	if (dbitype!=DBI_TYPE_ERROR) SQAPI->throwerror(v,"encountered unknown database field type") ;
					SQAPI->pushbool(v,false) ;
					DenObject ret(v) ;
					return ret ;
				 }

		}
	}
	//should never reach this point
}


static DenObject GetField(HSQUIRRELVM v,dbi_result res,const char *fieldname)
{
	unsigned int idx=dbi_result_get_field_idx(res,fieldname) ;
	if (idx) return GetField(v,res,idx) ;
	SQAPI->pushbool(v,false) ;
	DenObject falseobj(v) ; //if field does not exist, return a "false" bool object (we can't throw an error here :-/)
	SQAPI->throwerror(v,"column name") ;
	return falseobj ;
}

SQInteger DBIResult::SQ_constructor( HSQUIRRELVM v )
{
	SQAPI->throwerror(v,"no manual construction") ;
	return SQ_ERROR ;
}

//returns null on empty result, false on error
SQInteger DBIResult::SQ_column( HSQUIRRELVM v )
{
	_DENAPI::stack sa(v) ;
	DBIResult *pxThis=sa.GetInstancePtr<DBIResult>(1) ;
	if (!den_assert(pxThis,v,"internal error")) return SQ_ERROR ;
	if ( !den_assert( sa.GetParamCount()==2 && (sa.GetType(2)==OT_INTEGER||sa.GetType(2)==OT_STRING), v, "parameter errors" ) ) return SQ_ERROR ;
	if ( !pxThis->mResult ) return sa.ReturnNull() ; //no result present (wtf?)
	dbi_result pxres = pxThis->mResult ;
	unsigned short fieldtype = DBI_TYPE_ERROR ; 
	int idx = -1 ;
	const char *pst = NULL ;
	if ( sa.GetType( 2 )==OT_STRING )
	{
		pst=sa.GetString(2) ;
		fieldtype=dbi_result_get_field_type( pxres, pst) ;
	}
	else //if ( sa.GetType( 2 )==OT_INTEGER )
	{
		idx = sa.GetInt( 2 ) ;
		fieldtype=dbi_result_get_field_type_idx( pxres, idx)	 ;
	}
	DenObject retval(idx>=0?GetField(v,pxres,idx):GetField(v,pxres,pst)) ;
	return retval.GetType()==OT_BOOL? SQ_ERROR : sa.Return(retval) ;
	
	
}

SQInteger DBIResult::SQ_table( HSQUIRRELVM v )
{
	return SQ_row(v) ;
}


SQInteger DBIResult::SQ_row( HSQUIRRELVM v )
{
	_DENAPI::stack sa(v) ;
	DBIResult *pxThis=sa.GetInstancePtr<DBIResult>(1) ;
	if (!den_assert(pxThis && pxThis->mConn,v,"internal error")) return SQ_ERROR ;
	DenObject tab=pxThis->GetRowTable(v) ;
	if (tab.GetType()==OT_BOOL) return SQ_ERROR ;
	return sa.Return(tab) ;
}

SQInteger DBIResult::SQ_array( HSQUIRRELVM v )
{
	_DENAPI::stack sa(v) ;
	DBIResult *pxThis=sa.GetInstancePtr<DBIResult>(1) ;
	if (!den_assert(pxThis && pxThis->mConn  && pxThis->mResult,v,"internal error")) return SQ_ERROR ;
	size_t len=dbi_result_get_numrows(pxThis->mResult)-dbi_result_get_currow(pxThis->mResult) ;
	SQAPI->newarray(v,(SQInteger)len) ;
	SQInteger i=0 ;
	while (dbi_result_next_row(pxThis->mResult))
	{
		DenObject tab=pxThis->GetRowTable(v) ;
		if (tab.GetType()==OT_BOOL) { SQAPI->poptop(v); return SQ_ERROR ; }
		SQInteger t=SQAPI->gettop(v) ;
		SQAPI->pushinteger(v,i++) ;
		tab.Push(v) ;
		SQAPI->rawset(v,-3) ;
		SQAPI->settop(v,t) ;
	}
	return 1 ;
}

SQInteger DBIResult::SQ_next_row( HSQUIRRELVM v )
{
	_DENAPI::stack sa(v) ;
	DBIResult *pxThis=sa.GetInstancePtr<DBIResult>(1) ;
	if (!den_assert(pxThis && pxThis->mConn && pxThis->mResult,v,"internal error")) return SQ_ERROR ;
	return sa.Return((SQInteger)dbi_result_next_row(pxThis->mResult)) ;
}

DenObject DBIResult::GetRowTable( HSQUIRRELVM v )
{
	
	SQAPI->pushbool(v,false) ;
	DenObject FalseObj(v) ;
	SQAPI->newtable(v) ;
	unsigned int flds=dbi_result_get_numfields(mResult) ;
	for (int i=1;i<=flds;i++)
	{
		const char *key=dbi_result_get_field_name(mResult,i) ;
		if (!key) { SQAPI->throwerror(v,"index error") ; SQAPI->poptop(v) ; return FalseObj; }
		DenObject val=GetField(v,mResult,i) ;
		if (val.GetType()==OT_BOOL) { SQAPI->poptop(v) ; return FalseObj ; }
		int t=SQAPI->gettop(v) ;
		SQAPI->pushstring(v,key,-1) ;
		val.Push(v) ;
		SQAPI->newslot(v,-3,false) ;
		SQAPI->settop(v,t) ;
	}
	DenObject ret(v) ;
	return ret ;
}
SQInteger DBIResult::SQ_Error( HSQUIRRELVM v )
{
	_DENAPI::stack sa(v) ;
	DBIResult *pxThis=sa.GetInstancePtr<DBIResult>(1) ;
	if (!den_assert(pxThis,v,"internal error")) return SQ_ERROR ;
	return sa.Return(pxThis->Error()) ;
}

SQInteger DBIResult::SQ_first_row( HSQUIRRELVM v )
{
	_DENAPI::stack sa(v) ;
	DBIResult *pxThis=sa.GetInstancePtr<DBIResult>(1) ;
	if (!den_assert(pxThis && pxThis->mConn && pxThis->mResult,v,"internal error")) return SQ_ERROR ;
	return sa.Return((SQInteger)dbi_result_first_row(pxThis->mResult)) ;
}

DBICon::DBICon( const char *drivername,const char*hostorpath,const char*username,const char*passw,const char*dbnameorfile,int port/*=0*/,const char *IanaCode/*=NULL*/ )
:mConn(NULL)

{
	m_Mutex.Lock() ;
	if (!miInstcounter) 
	{	//first instance, initialize") ;
		dbi_initialize (NULL) ;

		dbi_driver t=NULL ;
		/*while (t=dbi_driver_list(t))
		{
			const char *s=dbi_driver_get_name(t) ;
			LogDebug("Available database driver: %s",s) ;
		}
		*/
	}
	miInstcounter++ ;
	mConn=dbi_conn_new(drivername);
	m_errmsg="" ;
	
	if (!mConn) { m_errmsg="unable to load requested DBI driver" ; } 
	else {
		//LogNotice("Using database driver '%s' on %s db %s",type,host_or_path,dbname) ;
		dbi_conn_set_option_numeric(mConn,"port",port) ;

		dbi_conn_set_option(mConn,"host",hostorpath) ; //in case driver is NOT sqlite (which will ignore these...)
		dbi_conn_set_option(mConn,"username",username) ; //...
		dbi_conn_set_option(mConn,"password",passw) ; //...
		dbi_conn_set_option(mConn,"sqlite3_dbdir",hostorpath) ; //in case driver is sqlite, other drivers will ignore this
		dbi_conn_set_option(mConn,"dbname",dbnameorfile) ; //used by all known drivers
		if (port!=0) dbi_conn_set_option_numeric(mConn,"sqlite3_timeout",port) ; //timeout in milliseconds for concurring access JUST IN CASE driver is sqlite, other drivers will ignore this
		dbi_conn_error_handler(mConn,ErrorHandler,this) ;
		if (IanaCode)
		{
			const char *Iana=dbi_driver_encoding_from_iana(dbi_conn_get_driver(mConn),IanaCode) ;
			dbi_conn_set_option(mConn,"encoding",Iana) ;
		}

	}
	dbi_conn_connect(mConn) ; //we MUST connect here to avoid a crash when user does not connect manually before querying (which involves pinging which in turn will make mysqlclient 5.1.x or lower crash)
	m_Mutex.Unlock() ;
}

DBICon::~DBICon()
{
	
	m_Mutex.Lock() ;
	if (mConn) dbi_conn_close(mConn) ;
	if (!--miInstcounter) { dbi_shutdown() ; }
	m_Mutex.Unlock() ;
}

void DBICon::BindScriptMethods( HSQUIRRELVM v)
{
	DENBIND_FUNC( v, constructor );
	DENBIND_FUNC( v, connect) ;
	DENBIND_FUNC( v, close) ;
	DENBIND_FUNC( v, Error) ;
	DENBIND_FUNC( v, query) ;

	DENBIND_FUNC( v, LastAutoIncrement);
	DENBIND_FUNC( v, quote) ;
	DENBIND_FUNC( v, likequote) ;
	DENBIND_FUNC( v, LastAutoIncrement);
	DENBIND_FUNC( v, selectdatabase);
}

dbi_result DBICon::query( const SQChar *qry )
{
	
	if (!mConn) return NULL ;
	
	m_Mutex.Lock() ;
	UpdateErrorMsg(NULL) ;
	if (!dbi_conn_ping(mConn)) {
		dbi_conn_connect(mConn) ; //conn_ping must be protected by mutex, same as query and error
		//printf("dbi conn\r\n") ;
	}
	dbi_result res=NULL ;
	res=dbi_conn_query(mConn,qry) ;
	const char *mlerr=NULL ;
	int errorcode=dbi_conn_error(mConn,&mlerr) ;
	if (mlerr) { m_errmsg=mlerr ; 
		UpdateErrorMsg(mlerr) ; 
	}
	m_Mutex.Unlock() ;
	return res ;
}

SQInteger DBICon::SQ_constructor( HSQUIRRELVM v )
{
	_DENAPI::stack sa(v) ;
	if (!den_assert(sa.GetParamCount()>=6 && sa.GetParamCount()<9,v,"parameter count mismatch: (drivername,hostorpath,username,password,dbnameorfilename[,port[,IanaEncoding]])")) { return SQ_ERROR ;} ;
	if (!den_assert(sa.GetType(2)==OT_STRING,v,"1st parameter type error")) { return SQ_ERROR; } ;
	if (!den_assert(sa.GetType(3)==OT_STRING || sa.GetType(3)==OT_NULL,v,"2nd parameter type error")) { return SQ_ERROR; } ;
	if (!den_assert(sa.GetType(4)==OT_STRING || sa.GetType(4)==OT_NULL,v,"3rd parameter type error")) { return SQ_ERROR; } ;
	if (!den_assert(sa.GetType(5)==OT_STRING || sa.GetType(5)==OT_NULL,v,"4th parameter type error")) { return SQ_ERROR; } ;
	if (!den_assert(sa.GetType(6)==OT_STRING || sa.GetType(6)==OT_NULL,v,"5th parameter type error")) { return SQ_ERROR; } ;
	if (!den_assert(sa.GetParamCount()<7 || sa.GetType(7)==OT_INTEGER,v,"6th parameter type error")) { return SQ_ERROR; } ;
	if (!den_assert(sa.GetParamCount()<8 || sa.GetType(8)==OT_STRING,v,"7th parameter type error")) { return SQ_ERROR; } ;
	//const SQChar *fspec=sa.GetString(2) ;
	const char *drivername=sa.GetString(2) ;
	const char *hostorpath=NULL ;
	if (sa.GetType(3)==OT_STRING) hostorpath=sa.GetString(3) ;
	const char *username=NULL ;
	if (sa.GetType(4)==OT_STRING) username=sa.GetString(4) ;
	const char *passw=NULL ;
	if (sa.GetType(5)==OT_STRING) passw=sa.GetString(5) ;
	const char *dbnameorpath=NULL ;
	if (sa.GetType(6)==OT_STRING) dbnameorpath=sa.GetString(6) ;
	SQInteger port=0 ;
	if (sa.GetParamCount()>6)
	{
		port=sa.GetInt(7) ;
	}
	const char *ianacode=NULL ;
	if (sa.GetParamCount()>7)
	{
		ianacode=sa.GetString(8) ;
	}
	_DENAPI::BindToInstanceOnStack( v, new DBICon(drivername,hostorpath,username,passw,dbnameorpath,port,ianacode), 1 );
	return 0 ; //attention: constructors return 0 or SQ_ERROR, never 1
}

SQInteger DBICon::SQ_query( HSQUIRRELVM v )
{
	_DENAPI::stack sa(v) ;
	DBICon *pxThis=sa.GetInstancePtr<DBICon>(1) ;
	if (!den_assert(pxThis,v,"internal error")) return SQ_ERROR ;
	if (!den_assert(sa.GetParamCount()==2 && sa.GetType(2)==OT_STRING,v,"parameter mismatch")) return SQ_ERROR ;
	dbi_result res=pxThis->query(sa.GetString(2)) ;
	if (res) {
		DBIResult *obj=new DBIResult(pxThis,res,pxThis->m_errmsg) ;
		return sa.Return(obj) ;
	}
	return sa.ReturnNull() ;
}

SQInteger DBICon::SQ_Error( HSQUIRRELVM v )
{
	_DENAPI::stack sa(v) ;
	DBICon *pxThis=sa.GetInstancePtr<DBICon>(1) ;
	if (!den_assert(pxThis,v,"internal error")) return SQ_ERROR ;
	return sa.Return(pxThis->Error()) ;
}

SQInteger DBICon::SQ_quote( HSQUIRRELVM v )
{
	_DENAPI::stack sa(v) ;
	if (!den_assert(sa.GetParamCount()==2 && (sa.GetType(2)==OT_STRING || sa.GetType(2)==OT_INSTANCE),v,"parameter error")) return SQ_ERROR ;
	DBICon * pxThis=sa.GetInstancePtr<DBICon>(1) ;
	if (!pxThis) { SQAPI->throwerror(v,"internal error 1") ; return SQ_ERROR; }
	
	if (sa.GetType(2)==OT_STRING)
	{
		char *d = NULL ;
		dbi_conn_quote_string_copy(pxThis->mConn,sa.GetString(2),&d) ;
		if (!d) { SQAPI->throwerror(v,"internal error 2") ; return SQ_ERROR; }
		SQInteger r=sa.Return(d) ;
		free(d) ;
		return r ;
	}
	//try to quote any binary data to a string
	SQInteger bufsiz=0 ;
	SQChar *buf=DENAPI.GetAnyBuffer(v,2,&bufsiz) ;
	SQChar *torelease=NULL ;
	if (!buf) {
		buf=DENAPI.CopyAnybuffer(v,2,&bufsiz) ;
		torelease=buf ;
	}
	if (!den_assert(buf,v,"illegal parameter")) return SQ_ERROR ;
	unsigned char *out=NULL ;
	dbi_conn_quote_binary_copy(pxThis->mConn,(const unsigned char *)buf,bufsiz,&out) ;
	if (torelease) SQAPI->free(torelease,bufsiz) ;
	if (out) {
		SQInteger r=sa.Return((SQChar*)out) ;
		free(out) ;
		return r ;
	}
	SQAPI->throwerror(v,"internal error 3") ; 
	return SQ_ERROR ;
}


SQInteger DBICon::SQ_likequote( HSQUIRRELVM v )
{
	_DENAPI::stack sa(v) ;
	if (!den_assert(sa.GetParamCount()==2 && sa.GetType(2)==OT_STRING,v,"parameter mismatch")) return SQ_ERROR ;
	const SQChar *inp=sa.GetString(2) ;
	std::string str("'") ; 
	int len=2 ;	while (inp && *inp++) { len++ ;}
	inp=sa.GetString(2) ;
	str.reserve(len) ;
	const SQChar *cur=inp ;
	bool escaped=false ;
	while (*cur)
	{
		switch (*cur)
		{
			//case 0:	if (cur-inp) str.append(inp,cur-inp) ; str.append("\\0",2) ; inp=cur+1;break ;
		case '\'': if (cur-inp) str.append(inp,cur-inp) ; str.append("''",2) ; inp=cur+1;escaped=true;break ; 
			//case '"' :if (cur-inp) str.append(inp,cur-inp) ; str.append("\\\"",2) ; inp=cur+1;escaped=true;break ;
			//case '\b':if (cur-inp) str.append(inp,cur-inp) ; str.append("\\b",2) ; inp=cur+1;escaped=true;break ;
			//case '\n':if (cur-inp) str.append(inp,cur-inp) ; str.append("\\n",2) ; inp=cur+1;escaped=true;break ;
			//case '\r':if (cur-inp) str.append(inp,cur-inp) ; str.append("\\r'",2) ; inp=cur+1;escaped=true;break ;
			//case '\t':if (cur-inp) str.append(inp,cur-inp) ; str.append("\\t'",2) ; inp=cur+1;escaped=true; break ;
			//case 26:if (cur-inp) str.append(inp,cur-inp) ; str.append("\\Z",2) ; inp=cur+1;escaped=true; break ;
		case '%':if (cur-inp) str.append(inp,cur-inp) ; str.append("\\%",2) ; inp=cur+1;escaped=true; break ;
		case '_':if (cur-inp) str.append(inp,cur-inp) ; str.append("\\_",2) ; inp=cur+1 ; break ;
		}
		cur++ ;
	}
	if (cur-inp) str.append(inp,cur-inp) ;
	if (escaped) str.append("' ESCAPE '\\'") ; else str.append("'") ;
	return sa.Return(str.c_str()) ;
}

SQInteger DBICon::SQ_LastAutoIncrement( HSQUIRRELVM v )
{
	_DENAPI::stack sa(v) ;
	DBICon * pxThis=sa.GetInstancePtr<DBICon>(1) ;
	if (!den_assert(pxThis,v,"internal error")) return SQ_ERROR ;
#ifdef _SQ64
	return sa.Return((SQInteger)dbi_conn_sequence_last(pxThis->mConn,NULL)) ; //return SQInteger as it fits!
#else
	long long ret=dbi_conn_sequence_last(pxThis->mConn,NULL) ;
	if (ret<0x7FFFFFFFL) return sa.Return((SQInteger)ret) ; //return an integer if it fits into 32 bit as a positive number
	SQChar* buf=sq_getscratchpad(v,64) ;
#ifdef SQUNICODE
	swprintf(buf,63,"%lld",ret);
#else	
	snprintf(buf,64,"%lld",ret) ;
#endif
	return sa.Return(buf) ; //otherwise, return the 64 bit value as a string
#endif

	
}

SQInteger DBICon::SQ_selectdatabase( HSQUIRRELVM v )
{
	_DENAPI::stack sa(v) ;
	DBICon * pxThis=sa.GetInstancePtr<DBICon>(1) ;
	if (!den_assert(pxThis,v,"internal error")) return SQ_ERROR ;
	if (!den_assert(sa.GetParamCount()==2 && sa.GetType(2)==OT_STRING,v,"parameter mismatch")) return SQ_ERROR ;
	pxThis->m_Mutex.Lock() ;
	bool ret=(dbi_conn_select_db(pxThis->mConn,sa.GetString(2))==0) ;
	pxThis->m_Mutex.Unlock() ;
	return sa.Return(ret) ;
}

SQInteger DBICon::SQ_connect( HSQUIRRELVM v )
{
	_DENAPI::stack sa(v) ;
	DBICon * pxThis=sa.GetInstancePtr<DBICon>(1) ;
	if (!den_assert(pxThis,v,"internal error")) return SQ_ERROR ;
	return sa.Return(dbi_conn_connect(pxThis->mConn)==0) ;
	
	

}

SQInteger DBICon::SQ_close( HSQUIRRELVM v )
{
	_DENAPI::stack sa(v) ;
	DBICon * pxThis=sa.GetInstancePtr<DBICon>(1) ;
	if (!den_assert(pxThis,v,"internal error")) return SQ_ERROR ;
	dbi_conn_close(pxThis->mConn) ;
	return 0 ;
}

void DBICon::ErrorHandler( dbi_conn Conn, void *arg )
{
	//unnecessary stuff , we avoid asynchronity
	/*DBICon *pxThis=(DBICon *)arg ;
	if (pxThis) 
	{
		const char *pError=0 ;
		int errcod=dbi_conn_error(pxThis->mConn,&pError) ; //TODO: Fuck, we'd need a mutex here, just we're in some sort of callback where a mutex could lock up!
		pxThis->m_errmsg=pError ;	
	}*/
}
