#pragma once
#ifndef SQDBI_H
#define SQDBI_H
#include "../DenAPI/DenAPI.h"
#include "../DenAPI/convenience.h"
#include "dbi/dbi.h"
#include <string>
class dbiquery ;
class pxdbi ;
typedef void (*QueryCallback)(dbiquery *result, pxdbi *db) ; //callback prototype for asynchronous queries


void SQDBI_RegisterClasses(HSQUIRRELVM v) ; //call this to register SQDBI classes to Squirrel. Note: Stub/Does simply NOTHING under windows (libdbi not available!)
//Squirrel wrappers for pxdbi stuff
class DBICon ;


class DBIResult : public DenBound< DBIResult >
{
public:
	DBIResult(DBICon *con,dbi_result res,std::string &errmsg) ;
	virtual ~DBIResult() ;
	static const char* ClassName() { return "DBIResult"; }
	static void BindScriptMethods( HSQUIRRELVM ) ;
	SQInteger next_row() ;
	const SQChar *Error() { return m_errmsg.empty()?NULL:m_errmsg.c_str() ; } ;

private:
	
	static SQInteger SQ_constructor( HSQUIRRELVM v ) ;
	static SQInteger SQ_column( HSQUIRRELVM v ) ;
	static SQInteger SQ_table( HSQUIRRELVM v) ;
	static SQInteger SQ_row(HSQUIRRELVM v) ;
	static SQInteger SQ_array( HSQUIRRELVM v) ;
	static SQInteger SQ_next_row(HSQUIRRELVM v);
	static SQInteger SQ_first_row(HSQUIRRELVM v) ;
	static SQInteger SQ_Error(HSQUIRRELVM v);
	
	DenObject GetRowTable(HSQUIRRELVM v) ;
	dbi_result mResult ;
	DBICon *mConn ; //note all DenBound classes are additionally reference counted (APIImpl will make use of that)
	int miFirstResult ;
	bool mbHasFirstResult ;
	friend class DBICon ;
	std::string m_errmsg ;
};

class DBICon : public DenBound< DBICon>
{
public:
	DBICon(const char *drivername,const char*hostorpath,const char*username,const char*passw,const char*dbnameorfile,int port=0,const char *IanaCode=NULL) ;
	virtual ~DBICon() ;
	static const char* ClassName() { return "DBICon"; }
	static void BindScriptMethods( HSQUIRRELVM ) ;
	dbi_result query(const SQChar *qry) ;
	const SQChar *Error() { return m_errmsg.empty()?NULL:m_errmsg.c_str() ; } ;
	std::string quote(const SQChar *) ;
	std::string likequote(const SQChar *) ;
	SQInteger close(void) ;
	inline void UpdateErrorMsg(const SQChar *err) { if (err) m_errmsg=err ; else m_errmsg.clear() ;}

private:
	static SQInteger SQ_constructor( HSQUIRRELVM v ) ;
	static SQInteger SQ_query(HSQUIRRELVM v) ;
	static SQInteger SQ_Error(HSQUIRRELVM v) ;
	static SQInteger SQ_quote(HSQUIRRELVM v) ;
	static SQInteger SQ_likequote(HSQUIRRELVM v);
	static SQInteger SQ_LastAutoIncrement(HSQUIRRELVM v) ;
	static SQInteger SQ_selectdatabase(HSQUIRRELVM v);
	static SQInteger SQ_connect(HSQUIRRELVM v);
	static SQInteger SQ_close(HSQUIRRELVM v);

	static void ErrorHandler(dbi_conn Conn, void *arg);
	dbi_conn mConn ;	
	static SQInteger miInstcounter ; //LibDBI does not support instances until v0.9, and many distributions out there still use 0.8x so we need to watch  out ourselves
	den_mutex m_Mutex ; //....which requires ping, query and error to be protected by a mutex
	std::string m_errmsg ;
};



#endif